{% extends "base.html" %}

{% block title %}IceProd{% end %}

{% block page_title %}IceCube Production System{% end %}

{% block body %}
<h4>Running grid jobs since 2007</h4>

<h2 style="margin-top:1em">Active Datasets: <a href="/dataset">here</a></h2>

<h3 style="margin-top:2em"><a href="/submit">Submit a new dataset</a></h3>
{% end %}
