.macro_box
{
    background: #fff;
    padding-top: 5px;
    padding-bottom: 10px;
}
.macro_box li
{
    margin-top: 10px;
    margin-left: 5px;
    margin-right: 5px;
    padding-top: 5px;
    border-top: solid #999 1px;
}
.macro_box h4
{
    font-weight: bold;
    font-size: 14px;
}
.macro_desc
{
    font-size: 13px;
    margin-top: 5px;
}

#submit_box
{
    font-family: monospace;
    white-space: pre;
}
.submit_text
{
    line-height: 1.5;
    margin-right: 10px;
}
/*
.table_form  { display: table;      }
.table_form	p     { display: table-row;  }
.table_form	label {
    display: table-cell;
    padding-right: 1em;
}
.table_form	input {
    display: table-cell;
    width:400px;
}*/

span.hide {
    display:none;
}

#advanced_submit {
    padding: 4px;
}
#advanced_submit div.indent {
    margin: 2px 1em 0;
    border-top: 1px solid blue;
}
#advanced_submit div.indent.trays {
    border-top: 1px solid cyan;
}
#advanced_submit div.indent.modules {
    border-top: 1px solid green;
}
#advanced_submit .key {
    text-transform: capitalize;
}
#advanced_submit input.key {
    text-transform: none;
}
#advanced_submit .header {
    font-weight: bold;
}
#advanced_submit .path {
    display:none;
}
#advanced_submit .add {
    color: blue;
    cursor: pointer;
}
#advanced_submit .delete {
    background-color: #C72E10;
    color: white;
    font: 0.8em/0.8em Arial,serif;
    padding: 0 2px 2px;
    margin-left: 1em;
    vertical-align: middle;
    cursor: pointer;
    -moz-user-select: none;
    -webkit-user-select: none;
    -ms-user-select: none;
    user-select: none;
    display: none;
}
#advanced_submit :hover > .delete {
    display: inline-block;
}
#advanced_submit .indent > .delete {
    float: right;
    font: 1em Arial,serif;
    line-height: 0.8em;
}
#advanced_submit .null {
    display: inline-block;
}
#advanced_submit .editable {
    cursor: pointer;
}
#advanced_submit span.val {
    display: inline-block;
    min-height: 1em;
    min-width: 20px;
}
#advanced_submit div.empty {
    height: .2em;
    width: 0;
    border: 1px solid black;
}

#expert_submit div {
    vertical-align: top;
}
#expert_submit div.xml {
    border: 1px black solid;
    margin: 8px 0;
    padding: 8px;
}
#expert_submit div.xml+div.xml {
    margin-top: 12px;
}
#expert_submit h4.tag_title {
    text-transform: capitalize;
    margin: 0 0 .2em;
}
#expert_submit div.xml_attr {
    display: inline-block;
}
#expert_submit div.xml_attr_text {
    display: block;
}
#expert_submit div.xml_subtags, #expert_submit div.xml_attribs {
    margin: 3px 0;
}

#expert_submit div.xml {
    background:#EFF8FF;
    border-color:#FFF8EF;
}
#expert_submit div.xml div.xml {
    background:#C5E7FF;
    border-color:#FFE6C7;
}
#expert_submit div.xml div.xml div.xml {
    background:#97D4FF;
    border-color:#FFD297;
}
#expert_submit div.xml div.xml div.xml div.xml {
    background:#62BEFF;
    border-color:#FFBC66;
}
#expert_submit div.xml div.xml div.xml div.xml div.xml {
    background:#2CA7FF;
    border-color:#FFA735;
}
#expert_submit div.xml div.xml div.xml div.xml div.xml div.xml {
    background:#fff;
    border-color:#000;
}
.sidebox
{
    float: right;
    background:#fff;
    padding: 15px;
    margin: 10px;
}
.sidebox li
{
    line-height: 150%;
}

#submit_area textarea, #submit_area input, #submit_area button
{
    font-size: 13px;
    background-color: #fff;
    border: solid #777 1px;
    border: none;
    border-radius: .4em;
}
#submit_area button, #submit_area input
{
    padding: 5px 10px;
}
#submit_area button#submit_action
{
    background: #189abf;
    color: white;
}
#submit_area button#submit_action:active
{
    color: black;
    background-color: #fff;
}
#number_jobs
{
    width:5em;
}


#submit_area textarea
{
    width: 100%;
    resize: vertical;
    height: 4em;
    margin: .5em auto;
    display: block;
}
#submit_area #number_jobs,
#submit_area #group
{
    margin-right: 3em;
}
#submit_area #group
{
    font-size: 13px;
    background-color: #fff;
    border: none;
    padding: 5px 10px;
    appearance: none;
    -webkit-appearance: none;
    -moz-appearance: none;
}
form
{
    width: 100%;
}
#submit_area
{
    background-color: #dddfe5;
    border-radius: .4em;
    padding: .1em .5em .2em;

    background: #dddfe5; /* Old browsers */
    background: -moz-linear-gradient(-45deg, #dddfe5 0%, #c3cadd 100%); /* FF3.6-15 */
    background: -webkit-linear-gradient(-45deg, #dddfe5 0%,#c3cadd 100%); /* Chrome10-25,Safari5.1-6 */
    background: linear-gradient(135deg, #dddfe5 0%,#c3cadd 100%); /* W3C, IE10+, FF16+, Chrome26+, Opera12+, Safari7+ */
}
#submit_area .submit_contents
{
    padding: .2em .5em;
}

.tab_bar
{
    border-radius: 5px 5px 0 0;
    background-color: rgba(0, 0, 0, .1);
}
.tab_bar li
{
    display: inline;
}
#submit_area .tab_bar button
{
    border-radius: 0;
    background: none;
    margin: 0;
    padding: 10px;
}
#submit_area .tab_bar button:hover
{
    cursor: pointer;
}
#submit_area .tab_bar li:first-child button
{
    border-radius: 5px 0 0 0;
}
#submit_area .tab_bar button.active
{
    background-color: #dddfe5;
}

#error {
    background: #ff9999;
    margin: .5em 0;
    text-align: center;
    line-height: 1.3em;
}

